from lib.core.entities.base import AttachmentEntity
from lib.core.entities.base import BaseEntity as TmpBaseEntity
from lib.core.entities.base import ProjectEntity
from lib.core.entities.base import SettingEntity
from lib.core.entities.integrations import IntegrationEntity
from lib.core.entities.items import DocumentEntity
from lib.core.entities.items import Entity
from lib.core.entities.items import TmpImageEntity
from lib.core.entities.items import VideoEntity
from lib.core.entities.project_entities import AnnotationClassEntity
from lib.core.entities.project_entities import BaseEntity
from lib.core.entities.project_entities import ConfigEntity
from lib.core.entities.project_entities import FolderEntity
from lib.core.entities.project_entities import ImageEntity
from lib.core.entities.project_entities import ImageInfoEntity
from lib.core.entities.project_entities import MLModelEntity
from lib.core.entities.project_entities import S3FileEntity
from lib.core.entities.project_entities import TeamEntity
from lib.core.entities.project_entities import UserEntity
from lib.core.entities.project_entities import WorkflowEntity
from superannotate_schemas.schemas.internal.document import DocumentAnnotation
from superannotate_schemas.schemas.internal.pixel import PixelAnnotation
from superannotate_schemas.schemas.internal.vector import VectorAnnotation
from superannotate_schemas.schemas.internal.video import VideoAnnotation
from superannotate_schemas.schemas.internal.video import (
    VideoAnnotation as VideoExportAnnotation,
)
# from lib.core.entities.project_entities import ProjectEntity

__all__ = [
    # base
    "SettingEntity",
    # items
    "TmpImageEntity",
    "BaseEntity",
    "TmpBaseEntity",
    "Entity",
    "VideoEntity",
    "DocumentEntity",
    # Utils
    "AttachmentEntity",
    # project
    "ProjectEntity",
    "ConfigEntity",
    "WorkflowEntity",
    "FolderEntity",
    "ImageEntity",
    "ImageInfoEntity",
    "S3FileEntity",
    "AnnotationClassEntity",
    "UserEntity",
    "TeamEntity",
    "MLModelEntity",
    "IntegrationEntity",
    # annotations
    "DocumentAnnotation",
    "VideoAnnotation",
    "VectorAnnotation",
    "PixelAnnotation",
    "VideoExportAnnotation",
]



import functools
import sys
from inspect import signature

from lib import get_default_controller
from mixpanel import Mixpanel
from superannotate.logger import get_default_logger
from version import __version__

from .utils import parsers

logger = get_default_logger()


def get_mp_instance() -> Mixpanel:
    if "api.annotate.online" in get_default_controller()._backend_url:
        return Mixpanel("ca95ed96f80e8ec3be791e2d3097cf51")
    return Mixpanel("e741d4863e7e05b1a45833d01865ef0d")


def get_default(team_name, user_id, project_name=None):
    return {
        "SDK": True,
        "Paid": True,
        "Team": team_name,
        "Team Owner": user_id,
        "Project Name": project_name,
        "Project Role": "Admin",
        "Version": __version__,
    }


class Trackable:
    TEAM_DATA = None
    INITIAL_EVENT = {"event_name": "SDK init", "properties": {}}
    INITIAL_LOGGED = False

    def __init__(self, function, initial=False):
        self.function = function
        self._success = False
        self._initial = initial
        if initial:
            self.track()
        functools.update_wrapper(self, function)

    @staticmethod
    def extract_arguments(function, *args, **kwargs) -> dict:
        bound_arguments = signature(function).bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        return dict(bound_arguments.arguments)

    @property
    def team(self):
        return get_default_controller().get_team()

    @staticmethod
    def default_parser(function_name: str, kwargs: dict):
        properties = {}
        for key, value in kwargs:
            if isinstance(value, (str, int, float, bool, str)):
                properties[key] = value
            elif isinstance(value, (list, set, tuple)):
                properties[key] = len(value)
            elif isinstance(value, dict):
                properties[key] = value.keys()
            elif hasattr(value, "__len__"):
                properties[key] = len(value)
            else:
                properties[key] = str(value)
        return {
            "event_name": function_name,
            "properties": properties
        }

    def track(self, *args, **kwargs):
        try:
            function_name = self.function.__name__ if self.function else ""
            if self._initial:
                data = self.INITIAL_EVENT
                Trackable.INITIAL_LOGGED = True
                self._success = True
            else:
                data = {}
                arguments = self.extract_arguments(self.function, *args, **kwargs)
                if hasattr(parsers, function_name):
                    try:
                        data = getattr(parsers, function_name)(**arguments)
                    except Exception:
                        pass
                else:
                    data = self.default_parser(function_name, arguments)
            event_name = data.get("event_name", )
            properties = data.get("properties", {})
            team_data = self.team.data
            user_id = team_data.creator_id
            team_name = team_data.name
            properties["Success"] = self._success
            default = get_default(
                team_name=team_name,
                user_id=user_id,
                project_name=properties.get("project_name", None),
            )
            properties.pop("project_name", None)
            properties = {**default, **properties}

            if "pytest" not in sys.modules:
                get_mp_instance().track(user_id, event_name, properties)
        except Exception:
            pass

    def __call__(self, *args, **kwargs):
        try:
            controller = get_default_controller()
            if controller:
                self.__class__.TEAM_DATA = controller.get_team()
                result = self.function(*args, **kwargs)
                self._success = True
            else:
                raise Exception(
                    "SuperAnnotate config file not found."
                    " Please provide correct config file location to sa.init(<path>) or use "
                    "CLI's superannotate init to generate default location config file."
                )
        except Exception as e:
            self._success = False
            logger.debug(str(e), exc_info=True)
            raise e
        else:
            return result
        finally:
            try:
                self.track(*args, **kwargs)
            except Exception:
                pass


if __name__ == "lib.app.mixp.decorators" and not Trackable.INITIAL_LOGGED:
    Trackable(None, initial=True)



from abc import ABC
from abc import abstractmethod

from lib.core.exceptions import AppValidationException
from lib.core.response import Response


class BaseUseCase(ABC):
    def __init__(self):
        self._response = Response()

    @abstractmethod
    def execute(self) -> Response:
        raise NotImplementedError

    def _validate(self):
        for name in dir(self):
            try:
                if name.startswith("validate_"):
                    method = getattr(self, name)
                    method()
            except AppValidationException as e:
                self._response.errors = e

    def is_valid(self):
        self._validate()
        return not self._response.errors


class BaseInteractiveUseCase(BaseUseCase):
    @property
    def response(self):
        return self._response

    @property
    def data(self):
        return self.response.data

    @abstractmethod
    def execute(self):
        raise NotImplementedError



import json
from pathlib import Path

import pandas as pd
import plotly.express as px
from lib.app.exceptions import AppException
from lib.core import DEPRICATED_DOCUMENT_VIDEO_MESSAGE
from superannotate.logger import get_default_logger


logger = get_default_logger()


def aggregate_image_annotations_as_df(
    project_root,
    include_classes_wo_annotations=False,
    include_comments=False,
    include_tags=False,
    folder_names=None,
):
    """Aggregate annotations as pandas dataframe from project root.

    :param project_root: export path of the project
    :type project_root: Pathlike (str or Path)
    :param include_classes_wo_annotations: enables inclusion of classes info
                                           that have no instances in annotations
    :type include_classes_wo_annotations: bool
    :param include_comments: enables inclusion of comments info as commentResolved column
    :type include_comments: bool
    :param include_tags: enables inclusion of tags info as tag column
    :type include_tags: bool
    :param folder_names: Aggregate the specified folders from project_root.
                                If None aggregate all folders in the project_root.
    :type folder_names: (list of str)

    :return: DataFrame on annotations with columns:
                                        "imageName", "instanceId",
                                        "className", "attributeGroupName", "attributeName", "type", "error", "locked",
                                        "visible", "trackingId", "probability", "pointLabels",
                                        "meta" (geometry information as string), "commentResolved", "classColor",
                                        "groupId", "imageWidth", "imageHeight", "imageStatus", "imagePinned",
                                        "createdAt", "creatorRole", "creationType", "creatorEmail", "updatedAt",
                                        "updatorRole", "updatorEmail", "tag", "folderName"
    :rtype: pandas DataFrame
    """

    json_paths = list(Path(str(project_root)).glob("*.json"))
    if (
        json_paths
        and "___pixel.json" not in json_paths[0].name
        and "___objects.json" not in json_paths[0].name
    ):
        raise AppException(DEPRICATED_DOCUMENT_VIDEO_MESSAGE)

    logger.info("Aggregating annotations from %s as pandas DataFrame", project_root)

    annotation_data = {
        "imageName": [],
        "imageHeight": [],
        "imageWidth": [],
        "imageStatus": [],
        "imagePinned": [],
        "instanceId": [],
        "className": [],
        "attributeGroupName": [],
        "attributeName": [],
        "type": [],
        "error": [],
        "locked": [],
        "visible": [],
        "trackingId": [],
        "probability": [],
        "pointLabels": [],
        "meta": [],
        "classColor": [],
        "groupId": [],
        "createdAt": [],
        "creatorRole": [],
        "creationType": [],
        "creatorEmail": [],
        "updatedAt": [],
        "updatorRole": [],
        "updatorEmail": [],
        "folderName": [],
        "imageAnnotator": [],
        "imageQA": [],
    }

    if include_comments:
        annotation_data["commentResolved"] = []
    if include_tags:
        annotation_data["tag"] = []

    classes_path = Path(project_root) / "classes" / "classes.json"
    if not classes_path.is_file():
        raise AppException(
            "SuperAnnotate classes file "
            + str(classes_path)
            + " not found. Please provide correct project export root"
        )
    classes_json = json.load(open(classes_path))
    class_name_to_color = {}
    class_group_name_to_values = {}
    for annotation_class in classes_json:
        name = annotation_class["name"]
        color = annotation_class["color"]
        class_name_to_color[name] = color
        class_group_name_to_values[name] = {}
        for attribute_group in annotation_class["attribute_groups"]:
            class_group_name_to_values[name][attribute_group["name"]] = []
            for attribute in attribute_group["attributes"]:
                class_group_name_to_values[name][attribute_group["name"]].append(
                    attribute["name"]
                )

    def __append_annotation(annotation_dict):
        for annotation_key in annotation_data:
            if annotation_key in annotation_dict:
                annotation_data[annotation_key].append(annotation_dict[annotation_key])
            else:
                annotation_data[annotation_key].append(None)

    def __get_image_metadata(image_name, annotations):
        image_metadata = {"imageName": image_name}

        image_metadata["imageHeight"] = annotations["metadata"].get("height")
        image_metadata["imageWidth"] = annotations["metadata"].get("width")
        image_metadata["imageStatus"] = annotations["metadata"].get("status")
        image_metadata["imagePinned"] = annotations["metadata"].get("pinned")
        image_metadata["imageAnnotator"] = annotations["metadata"].get("annotatorEmail")
        image_metadata["imageQA"] = annotations["metadata"].get("qaEmail")
        return image_metadata

    def __get_user_metadata(annotation):
        annotation_created_at = pd.to_datetime(annotation.get("createdAt"))
        annotation_created_by = annotation.get("createdBy")
        annotation_creator_email = None
        annotation_creator_role = None
        if annotation_created_by:
            annotation_creator_email = annotation_created_by.get("email")
            annotation_creator_role = annotation_created_by.get("role")
        annotation_creation_type = annotation.get("creationType")
        annotation_updated_at = pd.to_datetime(annotation.get("updatedAt"))
        annotation_updated_by = annotation.get("updatedBy")
        annotation_updator_email = None
        annotation_updator_role = None
        if annotation_updated_by:
            annotation_updator_email = annotation_updated_by.get("email")
            annotation_updator_role = annotation_updated_by.get("role")
        user_metadata = {
            "createdAt": annotation_created_at,
            "creatorRole": annotation_creator_role,
            "creatorEmail": annotation_creator_email,
            "creationType": annotation_creation_type,
            "updatedAt": annotation_updated_at,
            "updatorRole": annotation_updator_role,
            "updatorEmail": annotation_updator_email,
        }
        return user_metadata

    annotations_paths = []

    if folder_names is None:
        project_dir_content = Path(project_root).glob("*")
        for entry in project_dir_content:
            if entry.is_file() and entry.suffix == ".json":
                annotations_paths.append(entry)
            elif entry.is_dir() and entry.name != "classes":
                annotations_paths.extend(list(entry.rglob("*.json")))
    else:
        for folder_name in folder_names:
            annotations_paths.extend(
                list((Path(project_root) / folder_name).rglob("*.json"))
            )

    if not annotations_paths:
        logger.warning(f"Could not find annotations in {project_root}.")
    if len(list(Path(project_root).rglob("*___objects.json"))) > 0:
        type_postfix = "___objects.json"
    else:
        type_postfix = "___pixel.json"
    for annotation_path in annotations_paths:
        annotation_json = json.load(open(annotation_path))
        parts = annotation_path.name.split(type_postfix)
        if len(parts) != 2:
            continue
        image_name = parts[0]
        image_metadata = __get_image_metadata(image_name, annotation_json)
        annotation_instance_id = 0
        if include_comments:
            for annotation in annotation_json["comments"]:
                comment_resolved = annotation["resolved"]
                comment_meta = {
                    "x": annotation["x"],
                    "y": annotation["y"],
                    "comments": annotation["correspondence"],
                }
                annotation_dict = {
                    "type": "comment",
                    "meta": comment_meta,
                    "commentResolved": comment_resolved,
                }
                user_metadata = __get_user_metadata(annotation)
                annotation_dict.update(user_metadata)
                annotation_dict.update(image_metadata)
                __append_annotation(annotation_dict)
        if include_tags:
            for annotation in annotation_json["tags"]:
                annotation_dict = {"type": "tag", "tag": annotation}
                annotation_dict.update(image_metadata)
                __append_annotation(annotation_dict)
        for annotation in annotation_json["instances"]:
            annotation_type = annotation.get("type", "mask")
            annotation_class_name = annotation.get("className")
            if (
                annotation_class_name is None
                or annotation_class_name not in class_name_to_color
            ):
                logger.warning(
                    "Annotation class %s not found in classes json. Skipping.",
                    annotation_class_name,
                )
                continue
            annotation_class_color = class_name_to_color[annotation_class_name]
            annotation_group_id = annotation.get("groupId")
            annotation_locked = annotation.get("locked")
            annotation_visible = annotation.get("visible")
            annotation_tracking_id = annotation.get("trackingId")
            annotation_meta = None
            if annotation_type in ["bbox", "polygon", "polyline", "cuboid"]:
                annotation_meta = {"points": annotation["points"]}
            elif annotation_type == "point":
                annotation_meta = {"x": annotation["x"], "y": annotation["y"]}
            elif annotation_type == "ellipse":
                annotation_meta = {
                    "cx": annotation["cx"],
                    "cy": annotation["cy"],
                    "rx": annotation["rx"],
                    "ry": annotation["ry"],
                    "angle": annotation["angle"],
                }
            elif annotation_type == "mask":
                annotation_meta = {"parts": annotation["parts"]}
            elif annotation_type == "template":
                annotation_meta = {
                    "connections": annotation["connections"],
                    "points": annotation["points"],
                }
            annotation_error = annotation.get("error")
            annotation_probability = annotation.get("probability")
            annotation_point_labels = annotation.get("pointLabels")
            attributes = annotation.get("attributes")
            user_metadata = __get_user_metadata(annotation)
            folder_name = None
            if annotation_path.parent != Path(project_root):
                folder_name = annotation_path.parent.name
            num_added = 0
            if not attributes:
                annotation_dict = {
                    "imageName": image_name,
                    "instanceId": annotation_instance_id,
                    "className": annotation_class_name,
                    "type": annotation_type,
                    "locked": annotation_locked,
                    "visible": annotation_visible,
                    "trackingId": annotation_tracking_id,
                    "meta": annotation_meta,
                    "error": annotation_error,
                    "probability": annotation_probability,
                    "pointLabels": annotation_point_labels,
                    "classColor": annotation_class_color,
                    "groupId": annotation_group_id,
                    "folderName": folder_name,
                }
                annotation_dict.update(user_metadata)
                annotation_dict.update(image_metadata)
                __append_annotation(annotation_dict)
                num_added = 1
            else:
                for attribute in attributes:
                    attribute_group = attribute.get("groupName")
                    attribute_name = attribute.get("name")
                    if (
                        attribute_group
                        not in class_group_name_to_values[annotation_class_name]
                    ):
                        logger.warning(
                            "Annotation class group %s not in classes json. Skipping.",
                            attribute_group,
                        )
                        continue
                    if (
                        attribute_name
                        not in class_group_name_to_values[annotation_class_name][
                            attribute_group
                        ]
                    ):
                        logger.warning(
                            "Annotation class group value %s not in classes json. Skipping.",
                            attribute_name,
                        )
                        continue
                    annotation_dict = {
                        "imageName": image_name,
                        "instanceId": annotation_instance_id,
                        "className": annotation_class_name,
                        "attributeGroupName": attribute_group,
                        "attributeName": attribute_name,
                        "type": annotation_type,
                        "locked": annotation_locked,
                        "visible": annotation_visible,
                        "trackingId": annotation_tracking_id,
                        "meta": annotation_meta,
                        "error": annotation_error,
                        "probability": annotation_probability,
                        "pointLabels": annotation_point_labels,
                        "classColor": annotation_class_color,
                        "groupId": annotation_group_id,
                        "folderName": folder_name,
                    }
                    annotation_dict.update(user_metadata)
                    annotation_dict.update(image_metadata)
                    __append_annotation(annotation_dict)
                    num_added += 1

            if num_added > 0:
                annotation_instance_id += 1

    df = pd.DataFrame(annotation_data)

    # Add classes/attributes w/o annotations
    if include_classes_wo_annotations:
        for class_meta in classes_json:
            annotation_class_name = class_meta["name"]
            annotation_class_color = class_meta["color"]

            if annotation_class_name not in df["className"].unique():
                __append_annotation(
                    {
                        "className": annotation_class_name,
                        "classColor": annotation_class_color,
                    }
                )
                continue

            class_df = df[df["className"] == annotation_class_name][
                ["className", "attributeGroupName", "attributeName"]
            ]
            attribute_groups = class_meta["attribute_groups"]

            for attribute_group in attribute_groups:

                attribute_group_name = attribute_group["name"]

                attribute_group_df = class_df[
                    class_df["attributeGroupName"] == attribute_group_name
                ][["attributeGroupName", "attributeName"]]
                attributes = attribute_group["attributes"]
                for attribute in attributes:
                    attribute_name = attribute["name"]

                    if not (
                        attribute_name in attribute_group_df["attributeName"].unique()
                    ):
                        __append_annotation(
                            {
                                "className": annotation_class_name,
                                "classColor": annotation_class_color,
                                "attributeGroupName": attribute_group_name,
                                "attributeName": attribute_name,
                            }
                        )

        df = pd.DataFrame(annotation_data)

    df = df.astype({"probability": float})

    return df


def instance_consensus(inst_1, inst_2):
    """Helper function that computes consensus score between two instances:

    :param inst_1: First instance for consensus score.
    :type inst_1: shapely object
    :param inst_2: Second instance for consensus score.
    :type inst_2: shapely object

    """
    if inst_1.type == inst_2.type == "Polygon":
        intersect = inst_1.intersection(inst_2)
        union = inst_1.union(inst_2)
        score = intersect.area / union.area
    elif inst_1.type == inst_2.type == "Point":
        score = -1 * inst_1.distance(inst_2)
    else:
        raise NotImplementedError

    return score


def image_consensus(df, image_name, annot_type):
    """Helper function that computes consensus score for instances of a single image:

    :param df: Annotation data of all images
    :type df: pandas.DataFrame
    :param image_name: The image name for which the consensus score will be computed
    :type image_name: str
    :param annot_type: Type of annotation instances to consider. Available candidates are: ["bbox", "polygon", "point"]
    :type dataset_format: str

    """

    try:
        from shapely.geometry import Point, Polygon, box
    except ImportError:
        raise ImportError(
            "To use superannotate.benchmark or superannotate.consensus functions please install "
            "shapely package in Anaconda enviornment with # conda install shapely"
        )

    image_df = df[df["imageName"] == image_name]
    all_projects = list(set(df["folderName"]))
    column_names = [
        "creatorEmail",
        "imageName",
        "instanceId",
        "area",
        "className",
        "attributes",
        "folderName",
        "score",
    ]
    instance_id = 0
    image_data = {}
    for column_name in column_names:
        image_data[column_name] = []

    projects_shaply_objs = {}
    # generate shapely objects of instances
    for _, row in image_df.iterrows():
        if row["folderName"] not in projects_shaply_objs:
            projects_shaply_objs[row["folderName"]] = []
        inst_data = row["meta"]
        if annot_type == "bbox":
            inst_coords = inst_data["points"]
            x1, x2 = inst_coords["x1"], inst_coords["x2"]
            y1, y2 = inst_coords["y1"], inst_coords["y2"]
            inst = box(min(x1, x2), min(y1, y2), max(x1, x2), max(y1, y2))
        elif annot_type == "polygon":
            inst_coords = inst_data["points"]
            shapely_format = []
            for i in range(0, len(inst_coords) - 1, 2):
                shapely_format.append((inst_coords[i], inst_coords[i + 1]))
            inst = Polygon(shapely_format)
        elif annot_type == "point":
            inst = Point(inst_data["x"], inst_data["y"])
        if inst.is_valid:
            projects_shaply_objs[row["folderName"]].append(
                (inst, row["className"], row["creatorEmail"], row["attributes"])
            )
        else:
            logger.info(
                "Invalid %s instance occured, skipping to the next one.", annot_type
            )
    visited_instances = {}
    for proj, instances in projects_shaply_objs.items():
        visited_instances[proj] = [False] * len(instances)

    # match instances
    for curr_proj, curr_proj_instances in projects_shaply_objs.items():
        for curr_id, curr_inst_data in enumerate(curr_proj_instances):
            curr_inst, curr_class, _, _ = curr_inst_data
            if visited_instances[curr_proj][curr_id] == True:
                continue
            max_instances = []
            for other_proj, other_proj_instances in projects_shaply_objs.items():
                if curr_proj == other_proj:
                    max_instances.append((curr_proj, *curr_inst_data))
                    visited_instances[curr_proj][curr_id] = True
                else:
                    if annot_type in ["polygon", "bbox"]:
                        max_score = 0
                    else:
                        max_score = float("-inf")
                    max_inst_data = None
                    max_inst_id = -1
                    for other_id, other_inst_data in enumerate(other_proj_instances):
                        other_inst, other_class, _, _ = other_inst_data
                        if visited_instances[other_proj][other_id] == True:
                            continue
                        score = instance_consensus(curr_inst, other_inst)
                        if score > max_score and other_class == curr_class:
                            max_score = score
                            max_inst_data = other_inst_data
                            max_inst_id = other_id
                    if max_inst_data is not None:
                        max_instances.append((other_proj, *max_inst_data))
                        visited_instances[other_proj][max_inst_id] = True
            if len(max_instances) == 1:
                image_data["creatorEmail"].append(max_instances[0][3])
                image_data["attributes"].append(max_instances[0][4])
                image_data["area"].append(max_instances[0][1].area)
                image_data["imageName"].append(image_name)
                image_data["instanceId"].append(instance_id)
                image_data["className"].append(max_instances[0][2])
                image_data["folderName"].append(max_instances[0][0])
                image_data["score"].append(0)
            else:
                for curr_match_data in max_instances:
                    proj_cons = 0
                    for other_match_data in max_instances:
                        if curr_match_data[0] != other_match_data[0]:
                            score = instance_consensus(
                                curr_match_data[1], other_match_data[1]
                            )
                            proj_cons += 1.0 if score <= 0 else score
                    image_data["creatorEmail"].append(curr_match_data[3])
                    image_data["attributes"].append(curr_match_data[4])
                    image_data["area"].append(curr_match_data[1].area)
                    image_data["imageName"].append(image_name)
                    image_data["instanceId"].append(instance_id)
                    image_data["className"].append(curr_match_data[2])
                    image_data["folderName"].append(curr_match_data[0])
                    image_data["score"].append(proj_cons / (len(all_projects) - 1))
            instance_id += 1

    return image_data


def consensus_plot(consensus_df, *_, **__):
    plot_data = consensus_df.copy()

    # annotator-wise boxplot
    annot_box_fig = px.box(
        plot_data,
        x="creatorEmail",
        y="score",
        points="all",
        color="creatorEmail",
        color_discrete_sequence=px.colors.qualitative.Dark24,
    )
    annot_box_fig.show()

    # project-wise boxplot
    project_box_fig = px.box(
        plot_data,
        x="folderName",
        y="score",
        points="all",
        color="folderName",
        color_discrete_sequence=px.colors.qualitative.Dark24,
    )
    project_box_fig.show()

    # scatter plot of score vs area
    fig = px.scatter(
        plot_data,
        x="area",
        y="score",
        color="className",
        symbol="creatorEmail",
        facet_col="folderName",
        color_discrete_sequence=px.colors.qualitative.Dark24,
        hover_data={
            "className": False,
            "imageName": True,
            "folderName": False,
            "area": False,
            "score": False,
        },
    )
    fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
    fig.for_each_trace(lambda t: t.update(name=t.name.split("=")[-1]))
    fig.show()



import os
from pathlib import Path

import lib.core as constances
from lib.app.helpers import extract_project_folder
from lib.core.entities import IntegrationEntity
from lib.core.enums import ProjectType
from lib.infrastructure.controller import Controller


def get_project_name(project):
    project_name = ""
    if isinstance(project, dict):
        project_name = project["name"]
    if isinstance(project, str):
        if "/" in project:
            project_name = project.split("/")[0]
        else:
            project_name = project
    return project_name


def get_team_metadata(**kwargs):
    return {"event_name": "get_team_metadata", "properties": {}}


def invite_contributors_to_team(**kwargs):
    admin = kwargs.get("admin")
    if not admin:
        admin_value = False
    else:
        admin_value = admin
    return {
        "event_name": "invite_contributors_to_team",
        "properties": {"Admin": admin_value},
    }


def search_team_contributors(**kwargs):
    return {
        "event_name": "search_team_contributors",
        "properties": {
            "Email": bool(kwargs.get("email")),
            "Name": bool(kwargs.get("first_name")),
            "Surname": bool(kwargs.get("last_name")),
        },
    }


def search_projects(**kwargs):
    project = kwargs.get("name")
    return {
        "event_name": "search_projects",
        "properties": {
            "Metadata": bool(kwargs.get("return_metadata")),
            "project_name": get_project_name(project[0]) if project else None,
        },
    }


def create_project(**kwargs):
    project = kwargs["project_name"]
    project_type = kwargs["project_type"]
    return {
        "event_name": "create_project",
        "properties": {
            "Project Type": project_type,
            "project_name": get_project_name(project),
        },
    }


def create_project_from_metadata(**kwargs):
    project = kwargs.get("project_metadata")

    return {
        "event_name": "create_project_from_metadata",
        "properties": {"project_name": get_project_name(project)},
    }


def clone_project(**kwargs):
    project = kwargs.get("project_name")

    project_metadata = Controller.get_default().get_project_metadata(project).data["project"]
    project_type = ProjectType.get_name(project_metadata.type)

    return {
        "event_name": "clone_project",
        "properties": {
            "External": bool(
                project_metadata.upload_state == constances.UploadState.EXTERNAL.value
            ),
            "Project Type": project_type,
            "Copy Classes": bool(kwargs.get("copy_annotation_classes")
                                 ),
            "Copy Settings": bool(kwargs.get("copy_settings")),
            "Copy Workflow": bool(kwargs.get("copy_workflow")),
            "Copy Contributors": bool(kwargs.get("copy_contributors")
                                      ),
            "project_name": get_project_name(project),
        },
    }


def search_images(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "search_images",
        "properties": {
            "Annotation Status": bool(kwargs.get("annotation_status")
                                      ),
            "Metadata": bool(kwargs.get("return_metadata")),
            "project_name": get_project_name(project),
        },
    }


def upload_images_to_project(**kwargs):
    project = kwargs["project"]

    img_paths = kwargs.get("img_paths")
    return {
        "event_name": "upload_images_to_project",
        "properties": {
            "Image Count": len(img_paths) if img_paths else None,
            "Annotation Status": bool(kwargs.get("annotation_status")),
            "From S3": bool(kwargs.get("from_s3")),
            "project_name": get_project_name(project),
        },
    }


def upload_image_to_project(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "upload_image_to_project",
        "properties": {
            "Image Name": bool(kwargs.get("image_name")),
            "Annotation Status": bool(kwargs.get("annotation_status")
                                      ),
            "project_name": get_project_name(project),
        },
    }


def upload_video_to_project(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "upload_video_to_project",
        "properties": {
            "project_name": get_project_name(project),
            "FPS": bool(kwargs.get("target_fps")),
            "Start": bool(kwargs.get("start_time")),
            "End": bool(kwargs.get("end_time")),
        },
    }


def attach_image_urls_to_project(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "attach_image_urls_to_project",
        "properties": {
            "project_name": get_project_name(project),
            "Annotation Status": bool(kwargs.get("annotation_status")
                                      ),
        },
    }


def set_images_annotation_statuses(**kwargs):
    project = kwargs["project"]
    annotation_status = kwargs.get("annotation_status")
    image_names = kwargs["image_names"]
    return {
        "event_name": "set_images_annotation_statuses",
        "properties": {
            "project_name": get_project_name(project),
            "Image Count": len(image_names) if image_names else None,
            "Annotation Status": annotation_status,
        },
    }


def download_image_annotations(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "download_image_annotations",
        "properties": {"project_name": get_project_name(project)},
    }


def get_image_metadata(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "get_image_metadata",
        "properties": {"project_name": get_project_name(project)},
    }


def add_annotation_comment_to_image(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "add_annotation_comment_to_image",
        "properties": {"project_name": get_project_name(project)},
    }


def delete_annotation_class(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "delete_annotation_class",
        "properties": {"project_name": get_project_name(project)},
    }


def download_annotation_classes_json(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "download_annotation_classes_json",
        "properties": {"project_name": get_project_name(project)},
    }


def search_annotation_classes(**kwargs):
    project = kwargs["project"]
    name_prefix = kwargs.get("name_prefix")

    return {
        "event_name": "search_annotation_classes",
        "properties": {
            "project_name": get_project_name(project),
            "Prefix": bool(name_prefix),
        },
    }


def get_project_image_count(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "get_project_image_count",
        "properties": {"project_name": get_project_name(project)},
    }


def get_project_settings(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "get_project_settings",
        "properties": {"project_name": get_project_name(project)},
    }


def get_project_metadata(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "get_project_metadata",
        "properties": {"project_name": get_project_name(project)},
    }


def delete_project(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "delete_project",
        "properties": {"project_name": get_project_name(project)},
    }


def rename_project(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "rename_project",
        "properties": {"project_name": get_project_name(project)},
    }


def get_project_workflow(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "get_project_workflow",
        "properties": {"project_name": get_project_name(project)},
    }


def set_project_workflow(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "set_project_workflow",
        "properties": {"project_name": get_project_name(project)},
    }


def create_folder(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "create_folder",
        "properties": {"project_name": get_project_name(project)},
    }


def get_folder_metadata(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "get_folder_metadata",
        "properties": {"project_name": get_project_name(project)},
    }


def search_images_all_folders(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "search_images_all_folders",
        "properties": {
            "Annotation Status": bool(kwargs.get("annotation_status")
                                      ),
            "Metadata": bool(kwargs.get("return_metadata")),
            "project_name": get_project_name(project),
        },
    }


def download_model(**kwargs):
    model = kwargs["model"]
    return {
        "event_name": "download_model",
        "properties": {"model": model},
    }


def convert_project_type(**kwargs):
    return {
        "event_name": "convert_project_type",
        "properties": {},
    }


def convert_json_version(**kwargs):
    return {
        "event_name": "convert_json_version",
        "properties": {},
    }


def upload_image_annotations(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "upload_image_annotations",
        "properties": {
            "project_name": get_project_name(project),
            "Pixel": bool("mask" in kwargs),
        },
    }


def download_image(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "download_image",
        "properties": {
            "project_name": get_project_name(project),
            "Download Annotations": bool("include_annotations" in kwargs),
            "Download Fuse": bool("include_fuse" in kwargs),
            "Download Overlay": bool("include_overlay" in kwargs),
        },
    }


def copy_image(**kwargs):
    project = kwargs["source_project"]
    return {
        "event_name": "copy_image",
        "properties": {
            "project_name": get_project_name(project),
            "Copy Annotations": bool("include_annotations" in kwargs),
            "Copy Annotation Status": bool("copy_annotation_status" in kwargs),
            "Copy Pin": bool("copy_pin" in kwargs),
        },
    }


def run_prediction(**kwargs):
    project = kwargs["project"]
    project_name = get_project_name(project)
    res = Controller.get_default().get_project_metadata(project_name)
    project_metadata = res.data["project"]
    project_type = ProjectType.get_name(project_metadata.typy)
    image_list = kwargs["images_list"]
    return {
        "event_name": "run_prediction",
        "properties": {
            "Project Type": project_type,
            "Image Count": len(image_list) if image_list else None
        },
    }


def upload_videos_from_folder_to_project(**kwargs):
    folder_path = kwargs["folder_path"]
    glob_iterator = Path(folder_path).glob("*")
    return {
        "event_name": "upload_videos_from_folder_to_project",
        "properties": {"Video Count": sum(1 for _ in glob_iterator)},
    }


def export_annotation(**kwargs):
    dataset_format = kwargs["dataset_format"]
    project_type = kwargs["project_type"]
    if not project_type:
        project_type = "Vector"

    task = kwargs.get("task")
    if not task:
        task = "object_detection"
    return {
        "event_name": "export_annotation",
        "properties": {
            "Format": dataset_format,
            "Project Type": project_type,
            "Task": task,
        },
    }


def import_annotation(**kwargs):
    dataset_format = kwargs["dataset_format"]
    project_type = kwargs["project_type"]
    if not project_type:
        project_type = "Vector"
    task = kwargs.get("task")
    if not task:
        task = "object_detection"
    return {
        "event_name": "import_annotation",
        "properties": {
            "Format": dataset_format,
            "Project Type": project_type,
            "Task": task,
        },
    }


def move_images(**kwargs):
    project = kwargs["source_project"]
    project_name, folder_name = extract_project_folder(project)
    image_names = kwargs.get("image_names", False)
    if image_names is None:
        res = Controller.get_default().search_images(project_name, folder_name)
        image_names = res.data

    return {
        "event_name": "move_images",
        "properties": {
            "project_name": project_name,
            "Image Count": len(image_names),
            "Copy Annotations": bool("include_annotations" in kwargs),
            "Copy Annotation Status": bool("copy_annotation_status" in kwargs),
            "Copy Pin": bool("copy_pin" in kwargs),
        },
    }


def copy_images(**kwargs):
    project = kwargs["source_project"]
    project_name, folder_name = extract_project_folder(project)
    image_names = kwargs.get("image_names", False)
    if not image_names:
        res = Controller.get_default().search_images(project_name, folder_name)
        image_names = res.data
    return {
        "event_name": "copy_images",
        "properties": {
            "project_name": project_name,
            "Image Count": len(image_names),
            "Copy Annotations": bool("include_annotations" in kwargs),
            "Copy Annotation Status": bool("copy_annotation_status" in kwargs),
        },
    }


def consensus(**kwargs):
    folder_names = kwargs["folder_names"]
    image_list = kwargs["image_list"]
    annot_type = kwargs.get("annot_type")
    if not annot_type:
        annot_type = "bbox"
    show_plots = kwargs.get("show_plots")
    if not show_plots:
        show_plots = False
    return {
        "event_name": "consensus",
        "properties": {
            "Folder Count": len(folder_names),
            "Image Count": len(image_list) if image_list else None,
            "Annotation Type": annot_type,
            "Plot": show_plots,
        },
    }


def benchmark(**kwargs):
    folder_names = kwargs.get("folder_names")
    image_list = kwargs.get("image_list")
    annot_type = kwargs.get("annot_type")
    if not annot_type:
        annot_type = "bbox"
    show_plots = kwargs.get("show_plots")
    if not show_plots:
        show_plots = False

    return {
        "event_name": "benchmark",
        "properties": {
            "Folder Count": len(folder_names) if folder_names else None,
            "Image Count": len(image_list) if image_list else None,
            "Annotation Type": annot_type,
            "Plot": show_plots,
        },
    }


def upload_annotations_from_folder_to_project(**kwargs):
    project = kwargs["project"]
    project_name = get_project_name(project)
    res = Controller.get_default().get_project_metadata(project_name)
    project_metadata = res.data["project"]
    project_type = ProjectType.get_name(project_metadata.type)

    folder_path = kwargs["folder_path"]
    glob_iterator = Path(folder_path).glob("*.json")
    return {
        "event_name": "upload_annotations_from_folder_to_project",
        "properties": {
            "Annotation Count": sum(1 for _ in glob_iterator),
            "Project Type": project_type,
            "From S3": bool("from_s3_bucket" in kwargs),
        },
    }


def upload_preannotations_from_folder_to_project(**kwargs):
    project = kwargs["project"]

    project_name = get_project_name(project)
    res = Controller.get_default().get_project_metadata(project_name)
    project_metadata = res.data["project"]
    project_type = ProjectType.get_name(project_metadata.type)
    folder_path = kwargs["folder_path"]
    glob_iterator = Path(folder_path).glob("*.json")
    return {
        "event_name": "upload_preannotations_from_folder_to_project",
        "properties": {
            "Annotation Count": sum(1 for _ in glob_iterator),
            "Project Type": project_type,
            "From S3": bool("from_s3_bucket" in kwargs),
        },
    }


def upload_images_from_folder_to_project(**kwargs):
    folder_path = kwargs["folder_path"]
    recursive_subfolders = kwargs["recursive_subfolders"]
    extensions = kwargs["extensions"]
    if not extensions:
        extensions = constances.DEFAULT_IMAGE_EXTENSIONS
    exclude_file_patterns = kwargs["exclude_file_patterns"]
    if not exclude_file_patterns:
        exclude_file_patterns = constances.DEFAULT_FILE_EXCLUDE_PATTERNS

    paths = []
    for extension in extensions:
        if not recursive_subfolders:
            paths += list(Path(folder_path).glob(f"*.{extension.lower()}"))
            if os.name != "nt":
                paths += list(Path(folder_path).glob(f"*.{extension.upper()}"))
        else:
            paths += list(Path(folder_path).rglob(f"*.{extension.lower()}"))
            if os.name != "nt":
                paths += list(Path(folder_path).rglob(f"*.{extension.upper()}"))

    filtered_paths = []
    for path in paths:
        not_in_exclude_list = [x not in Path(path).name for x in exclude_file_patterns]
        if all(not_in_exclude_list):
            filtered_paths.append(path)

    return {
        "event_name": "upload_images_from_folder_to_project",
        "properties": {
            "Image Count": len(filtered_paths),
            "Custom Extentions": bool(kwargs["extensions"]),
            "Annotation Status": bool(kwargs.get("annotation_status")
                                      ),
            "From S3": bool(kwargs.get("from_s3_bucket")),
            "Custom Exclude Patters": bool(kwargs["exclude_file_patterns"]),
        },
    }


def prepare_export(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "prepare_export",
        "properties": {
            "project_name": get_project_name(project),
            "Folder Count": bool(kwargs.get("folder_names")),
            "Annotation Statuses": bool(kwargs.get("annotation_statuses")
                                        ),
            "Include Fuse": bool(kwargs.get("include_fuse")),
            "Only Pinned": bool(kwargs.get("only_pinned")),
        },
    }


def download_export(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "download_export",
        "properties": {
            "project_name": get_project_name(project),
            "to_s3_bucket": bool(kwargs.get("to_s3_bucket")),
        },
    }


def assign_images(**kwargs):
    project = kwargs["project"]
    project_name, folder_name = extract_project_folder(project)
    image_names = kwargs.get("image_names")
    user = kwargs.get("user")

    contributors = (
        Controller.get_default().get_project_metadata(project_name=project_name, include_contributors=True)
            .data["contributors"]
    )
    contributor = None
    for c in contributors:
        if c["user_id"] == user:
            contributor = c
    user_role = "ADMIN"
    if contributor["user_role"] == 3:
        user_role = "ANNOTATOR"
    if contributor["user_role"] == 4:
        user_role = "QA"
    is_root = True
    if folder_name:
        is_root = False

    return {
        "event_name": "assign_images",
        "properties": {
            "project_name": project_name,
            "Assign Folder": is_root,
            "Image Count": len(image_names) if image_names else None,
            "User Role": user_role,
        },
    }


def pin_image(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "pin_image",
        "properties": {
            "project_name": get_project_name(project),
            "Pin": bool("pin" in kwargs),
        },
    }


def set_image_annotation_status(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "set_image_annotation_status",
        "properties": {
            "project_name": get_project_name(project),
            "Annotation Status": bool("annotation_status" in kwargs),
        },
    }


def add_annotation_bbox_to_image(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "add_annotation_bbox_to_image",
        "properties": {
            "project_name": get_project_name(project),
            "Attributes": bool("annotation_class_attributes" in kwargs),
            "Error": bool("error" in kwargs),
        },
    }


def add_annotation_point_to_image(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "add_annotation_point_to_image",
        "properties": {
            "project_name": get_project_name(project),
            "Attributes": bool("annotation_class_attributes" in kwargs),
            "Error": bool("error" in kwargs),
        },
    }


def create_annotation_class(**kwargs):
    project = kwargs["project"]
    class_type = kwargs.get("class_type")

    return {
        "event_name": "create_annotation_class",
        "properties": {
            "project_name": get_project_name(project),
            "Attributes": bool("attribute_groups" in kwargs),
            "class_type": class_type if class_type else "object",
        },
    }


def create_annotation_classes_from_classes_json(**kwargs):
    project = kwargs["project"]
    return {
        "event_name": "create_annotation_classes_from_classes_json",
        "properties": {
            "project_name": get_project_name(project),
            "From S3": bool("from_s3_bucket" in kwargs),
        },
    }


def class_distribution(**kwargs):
    return {
        "event_name": "class_distribution",
        "properties": {"Plot": bool("visualize" in kwargs)},
    }


def share_project(**kwargs):
    project = kwargs["project_name"]

    user_role = kwargs.get("user_role")
    return {
        "event_name": "share_project",
        "properties": {
            "project_name": get_project_name(project),
            "User Role": user_role,
        },
    }


def set_project_default_image_quality_in_editor(**kwargs):
    project = kwargs["project"]

    image_quality_in_editor = kwargs.get("image_quality_in_editor")
    return {
        "event_name": "set_project_default_image_quality_in_editor",
        "properties": {
            "project_name": get_project_name(project),
            "Image Quality": image_quality_in_editor,
        },
    }


def get_exports(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "get_exports",
        "properties": {
            "project_name": get_project_name(project),
            "Metadata": bool("return_metadata" in kwargs),
        },
    }


def search_folders(**kwargs):
    project = kwargs["project"]

    return {
        "event_name": "search_folders",
        "properties": {
            "project_name": get_project_name(project),
            "Metadata": bool("return_metadata" in kwargs),
        },
    }


def aggregate_annotations_as_df(**kwargs):
    folder_names = kwargs.get("folder_names")
    if not folder_names:
        folder_names = []

    project_type = kwargs["project_type"]

    return {
        "event_name": "aggregate_annotations_as_df",
        "properties": {"Folder Count": len(folder_names), "Project Type": project_type},
    }


def delete_folders(**kwargs):
    project = kwargs["project"]
    folder_names = kwargs.get("folder_names")
    if not folder_names:
        folder_names = []
    return {
        "event_name": "delete_folders",
        "properties": {
            "project_name": get_project_name(project),
            "Folder Count": len(folder_names),
        },
    }


def delete_images(**kwargs):
    project = kwargs["project"]
    project_name, folder_name = extract_project_folder(project)

    image_names = kwargs.get("image_names", False)
    if not image_names:
        res = Controller.get_default().search_images(project_name, folder_name)
        image_names = res.data
    return {
        "event_name": "delete_images",
        "properties": {
            "project_name": project_name,
            "Image Count": len(image_names),
        },
    }


def unassign_folder(**kwargs):
    return {"event_name": "unassign_folder", "properties": {}}


def assign_folder(**kwargs):
    users = kwargs.get("users")
    return {"event_name": "assign_folder", "properties": {"User Count": len(users)}}


def unassign_images(**kwargs):
    image_names = kwargs.get("image_names")

    project = kwargs["project"]

    _, folder_name = extract_project_folder(project)
    is_root = True
    if folder_name:
        is_root = False

    return {
        "event_name": "unassign_images",
        "properties": {"Assign Folder": is_root, "Image Count": len(image_names)},
    }


def attach_video_urls_to_project(**kwargs):
    project = kwargs["project"]
    project_name, _ = extract_project_folder(project)
    return {
        "event_name": "attach_video_urls_to_project",
        "properties": {
            "project_name": project_name,
            "Annotation Status": bool(kwargs.get("annotation_status")),
        },
    }


def attach_document_urls_to_project(**kwargs):
    project = kwargs["project"]
    project_name, _ = extract_project_folder(project)
    return {
        "event_name": "attach_document_urls_to_project",
        "properties": {
            "project_name": project_name,
            "Annotation Status": bool(kwargs.get("annotation_status")),
        },
    }


def delete_annotations(**kwargs):
    return {"event_name": "delete_annotations", "properties": {}}


def validate_annotations(**kwargs):
    project_type = kwargs["project_type"]
    return {
        "event_name": "validate_annotations",
        "properties": {"Project Type": project_type},
    }


def add_contributors_to_project(**kwargs):
    user_role = kwargs.get("role")

    return {
        "event_name": "add_contributors_to_project",
        "properties": {"User Role": user_role},
    }


def get_annotations(**kwargs):
    project = kwargs["project"]
    items = kwargs["items"]

    return {
        "event_name": "get_annotations",
        "properties": {"Project": project, "items_count": len(items) if items else None},
    }


def get_annotations_per_frame(**kwargs):
    project = kwargs["project"]
    fps = kwargs["fps"]
    if not fps:
        fps = 1
    return {
        "event_name": "get_annotations_per_frame",
        "properties": {"Project": project, "fps": fps},
    }


def upload_priority_scores(**kwargs):
    scores = kwargs["scores"]
    return {
        "event_name": "upload_priority_scores",
        "properties": {"Score Count": len(scores) if scores else None},
    }


def get_integrations(**kwargs):
    return {
        "event_name": "get_integrations",
        "properties": {},
    }


def attach_items_from_integrated_storage(**kwargs):
    project = kwargs.get("project")
    project_name, _ = extract_project_folder(project)
    integration = kwargs.get("integration")
    folder_path = kwargs.get("folder_path")

    if isinstance(integration, str):
        integration = IntegrationEntity(name=integration)
    project = Controller.get_default().get_project_metadata(project_name).data["project"]
    return {
        "event_name": "attach_items_from_integrated_storage",
        "properties": {
            "project_type": ProjectType.get_name(project.type),
            "integration_name": integration.name,
            "folder_path": bool(folder_path),
        },
    }


def query(**kwargs):
    project = kwargs["project"]
    query_str = kwargs["query"]
    project_name, folder_name = extract_project_folder(project)
    project = Controller.get_default().get_project_metadata(project_name).data["project"]
    return {
        "event_name": "query_saqul",
        "properties": {
            "project_type": ProjectType.get_name(project.type),
            "query": query_str,
        },
    }


def get_item_metadata(**kwargs):
    project = kwargs["project"]
    project_name, _ = extract_project_folder(project)
    project = Controller.get_default().get_project_metadata(project_name).data["project"]
    return {
        "event_name": "get_item_metadata",
        "properties": {"project_type": ProjectType.get_name(project.type)},
    }


def search_items(**kwargs):
    project = kwargs["project"]
    name_contains = kwargs["name_contains"]
    annotation_status = kwargs["annotation_status"]
    annotator_email = kwargs["annotator_email"]
    qa_email = kwargs["qa_email"]
    recursive = kwargs["recursive"]
    project_name, folder_name = extract_project_folder(project)
    project = Controller.get_default().get_project_metadata(project_name).data["project"]
    return {
        "event_name": "search_items",
        "properties": {
            "project_type": ProjectType.get_name(project.type),
            "query": query,
            "name_contains": len(name_contains) if name_contains else False,
            "annotation_status": annotation_status if annotation_status else False,
            "annotator_email": bool(annotator_email),
            "qa_email": bool(qa_email),
            "recursive": bool(recursive),
        },
    }


def move_items(**kwargs):
    project = kwargs["project"]
    project_name, _ = extract_project_folder(project)
    project = Controller.get_default().get_project_metadata(project_name).data["project"]
    items = kwargs["items"]
    return {
        "event_name": "move_items",
        "properties": {
            "project_type": ProjectType.get_name(project.type),
            "items_count": len(items) if items else None
        },
    }


def copy_items(**kwargs):
    project = kwargs["project"]
    project_name, _ = extract_project_folder(project)
    project = Controller.get_default().get_project_metadata(project_name).data["project"]
    items = kwargs["items"]
    return {
        "event_name": "copy_items",
        "properties": {
            "project_type": ProjectType.get_name(project.type),
            "items_count": len(items) if items else None,
            "include_annotations": kwargs["include_annotations"]
        },
    }


def attach_items(**kwargs):
    project = kwargs["project"]
    project_name, _ = extract_project_folder(project)
    project = Controller.get_default().get_project_metadata(project_name).data["project"]
    attachments = kwargs["attachments"]
    return {
        "event_name": "copy_items",
        "properties": {
            "project_type": ProjectType.get_name(project.type),
            "attachments": "scv" if isinstance(attachments, (str, Path)) else "dict",
            "annotation_status": kwargs["annotation_status"]
        },
    }



from abc import ABC
from typing import Any
from typing import List
from typing import Set
from typing import Union

import superannotate.lib.core as constance
from pydantic import BaseModel
from superannotate.lib.core.entities import BaseEntity
from superannotate.lib.core.entities import ImageEntity
from superannotate.lib.core.entities import ProjectEntity


class BaseSerializer(ABC):
    def __init__(self, entity: BaseEntity):
        self._entity = entity

    @staticmethod
    def _fill_enum_values(data: dict):
        if isinstance(data, dict):
            for key, value in data.items():
                if hasattr(value, "_type") and value._type == "titled_enum":
                    data[key] = value.__doc__
        return data

    def serialize(
            self, fields: List[str] = None, by_alias: bool = True, flat: bool = False, exclude: Set[str] = None
    ):
        return self._fill_enum_values(
            self._serialize(self._entity, fields, by_alias, flat, exclude=exclude)
        )

    def serialize_item(
            self,
            data: Any,
            fields: Union[List[str], Set[str]] = None,
            by_alias: bool = False,
            flat: bool = False,
    ):
        return self._fill_enum_values(self._serialize(data, fields, by_alias, flat))

    @staticmethod
    def _serialize(
            entity: Any,
            fields: List[str] = None,
            by_alias: bool = False,
            flat: bool = False,
            exclude: Set[str] = None
    ):
        if isinstance(entity, dict):
            return entity
        if isinstance(entity, BaseModel):
            if fields:
                fields = set(fields)
                if len(fields) == 1:
                    if flat:
                        return entity.dict(include=fields, by_alias=by_alias, exclude=exclude)[
                            next(iter(fields))
                        ]
                    return entity.dict(include=fields, by_alias=by_alias, exclude=exclude)
                return entity.dict(include=fields, by_alias=by_alias, exclude=exclude)
            return entity.dict(by_alias=by_alias, exclude=exclude)
        return entity.to_dict()


    @classmethod
    def serialize_iterable(
            cls,
            data: List[Any],
            fields: Union[List[str], Set[str]] = None,
            by_alias: bool = False,
            flat: bool = False,
    ) -> List[Any]:
        serialized_data = []
        for i in data:
            serialized_data.append(
                cls._fill_enum_values(cls._serialize(i, fields, by_alias, flat))
            )
        return serialized_data


class UserSerializer(BaseSerializer):
    def serialize(self):
        data = super().serialize()
        data["user_role"] = constance.UserRole[data["user_role"]].name
        return data


class TeamSerializer(BaseSerializer):
    def serialize(self):
        data = super().serialize()
        users = []
        for user in data["users"]:
            user["user_role"] = constance.UserRole.get_name(user["user_role"])
            users.append(user)
        data["users"] = users
        for user in data["pending_invitations"]:
            user["user_role"] = constance.UserRole.get_name(user["user_role"])
        return data


class ProjectSerializer(BaseSerializer):
    DEFAULT_EXCLUDE_SET = {"sync_status", "unverified_users"}

    def serialize(self, fields: List[str] = None, by_alias: bool = False, flat: bool = False, exclude: Set[str] = None):
        if exclude:
            exclude = exclude.union(self.DEFAULT_EXCLUDE_SET)
        data = super().serialize(fields, by_alias, flat, exclude)
        if data.get("settings"):
            data["settings"] = [SettingsSerializer(setting).serialize() for setting in data["settings"]]
        data["type"] = constance.ProjectType.get_name(data["type"])
        if data.get("status"):
            data["status"] = constance.ProjectStatus.get_name(data["status"])
        else:
            data["status"] = "Undefined"
        if data.get("upload_state"):
            data["upload_state"] = constance.UploadState(data["upload_state"]).name
        if data.get("users"):
            for contributor in data["users"]:
                contributor["user_role"] = constance.UserRole.get_name(
                    contributor["user_role"]
                )
        return data


class FolderSerializer(BaseSerializer):
    def serialize(self):
        data = super().serialize()
        del data["is_root"]
        return data


class ImageSerializer(BaseSerializer):
    def serialize(self):
        data = super().serialize()
        data["annotation_status"] = constance.AnnotationStatus.get_name(
            data["annotation_status"]
        )
        return data

    def serialize_by_project(self, project: ProjectEntity):
        data = super().serialize()
        data = {
            "name": data.get("name"),
            "path": data.get("path"),
            "annotation_status": data.get("annotation_status"),
            "prediction_status": data.get("prediction_status"),
            "segmentation_status": data.get("segmentation_status"),
            "approval_status": data.get("approval_status"),
            "is_pinned": data.get("is_pinned"),
            "annotator_name": data.get("annotator_name"),
            "qa_name": data.get("qa_name"),
            "entropy_value": data.get("entropy_value"),
            "createdAt": data.get("createdAt"),
            "updatedAt": data.get("updatedAt"),
        }

        data["annotation_status"] = constance.AnnotationStatus.get_name(
            data["annotation_status"]
        )

        if project.upload_state == constance.UploadState.EXTERNAL.value:
            data["prediction_status"] = None
            data["segmentation_status"] = None
        else:
            if project.type == constance.ProjectType.VECTOR.value:
                data["prediction_status"] = constance.SegmentationStatus.get_name(
                    data["prediction_status"]
                )
                data["segmentation_status"] = None
            if project.type == constance.ProjectType.PIXEL.value:
                data["prediction_status"] = constance.SegmentationStatus.get_name(
                    data["prediction_status"]
                )
                data["segmentation_status"] = constance.SegmentationStatus.get_name(
                    data["segmentation_status"]
                )
            data["path"] = None
        return data

    @staticmethod
    def deserialize(data):
        if isinstance(data, list):
            return [ImageEntity(**image) for image in data]
        return ImageEntity(**data)


class SettingsSerializer(BaseSerializer):
    def serialize(self, fields: List[str] = None, by_alias: bool = True, flat: bool = False, exclude=None):
        data = super().serialize(fields, by_alias, flat, exclude)
        if data["attribute"] == "ImageQuality":
            data["value"] = constance.ImageQuality.get_name(data["value"])
        return data



"""
SA to COCO conversion methods
"""
import threading

from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from .coco_api import _area
from .coco_api import _merge
from .coco_api import _polytoMask
from .coco_api import _toBbox

logger = get_default_logger()


def sa_vector_to_coco_object_detection(
    make_annotation, image_commons, instances, id_generator
):
    annotations_per_image = []
    image_info = image_commons.image_info

    for instance in instances:
        if instance["type"] != "bbox":
            logger.warning(
                "Skipping '%s' type conversion during object_detection task",
                instance["type"],
            )
            continue

        anno_id = next(id_generator)
        category_id = instance["classId"]
        points = instance["points"]
        for key in points:
            points[key] = round(points[key], 2)
        bbox = (
            points["x1"],
            points["y1"],
            points["x2"] - points["x1"],
            points["y2"] - points["y1"],
        )
        polygons = bbox
        area = int((points["x2"] - points["x1"]) * points["y2"] - points["y1"])

        annotation = make_annotation(
            category_id, image_info["id"], bbox, polygons, area, anno_id
        )
        annotations_per_image.append(annotation)

    return image_info, annotations_per_image


def sa_vector_to_coco_instance_segmentation(
    make_annotation, image_commons, instances, id_generator
):
    annotations_per_image = []
    grouped_polygons = {}

    image_info = image_commons.image_info
    id_ = -1
    for instance in instances:
        if instance["type"] != "polygon":
            logger.warning(
                "Skipping '%s' type conversion during instance_segmentation task",
                instance["type"],
            )
            continue

        if instance["classId"] < 0:
            continue

        group_id = instance["groupId"]

        if group_id == 0:
            group_id += id_
            id_ -= 1

        category_id = instance["classId"]
        points = [round(point, 2) for point in instance["points"]]
        grouped_polygons.setdefault(group_id, {}).setdefault(category_id, []).append(
            points
        )

    for polygon_group in grouped_polygons.values():
        for cat_id, polygons in polygon_group.items():
            anno_id = next(id_generator)
            masks = _polytoMask(polygons, image_info["height"], image_info["width"])
            mask = _merge(masks)
            area = int(_area(mask))
            bbox = list(_toBbox(mask))
            annotation = make_annotation(
                cat_id, image_info["id"], bbox, polygons, area, anno_id
            )
            annotations_per_image.append(annotation)
    return (image_info, annotations_per_image)


def sa_vector_to_coco_keypoint_detection(
    jsons, id_generator, id_generator_anno, id_generator_img, make_image_info, total_num
):
    def __make_skeleton(template):
        res = [
            [connection["from"], connection["to"]]
            for connection in template["connections"]
        ]
        return res

    def __make_keypoints(template):
        int_dict = {int(key): value for key, value in template["pointLabels"].items()}

        res = [int_dict[i] for i in range(len(int_dict))]
        return res

    def __make_bbox(points):
        xs = [point["x"] for point in points]
        ys = [point["y"] for point in points]

        return [
            int(min(xs)),
            int(min(ys)),
            int(max(xs)) - int(min(xs)),
            int(max(ys)) - int(min(ys)),
        ]

    def __make_annotations(template, id_generator, cat_id, image_id):
        keypoints = []

        for point in template["points"]:
            keypoints += [round(point["x"], 2), round(point["y"], 2), 2]
        bbox = __make_bbox(template["points"])
        annotation = {
            "id": next(id_generator),
            "image_id": image_id,
            "iscrowd": 0,
            "bbox": bbox,
            "area": bbox[2] * bbox[3],
            "num_keypoints": len(template["points"]),
            "keypoints": keypoints,
            "category_id": cat_id,
        }

        return annotation

    template_names = set()
    categories = []
    annotations = []
    images = []

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(total_num, images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to COCO JSON format")
    tqdm_thread.start()
    for json_ in jsons:
        json_data = json_["instances"]
        for instance in json_data:
            if instance["type"] == "template" and "templateId" not in instance:
                logger.warning(
                    'There was a template with no "templateName". \
    This can happen if the template was deleted from superannotate.com. Ignoring this annotation'
                )
                continue

            if (
                instance["type"] != "template"
                or instance["templateId"] in template_names
            ):
                continue
            if instance["pointLabels"] == {}:
                instance["pointLabels"] = {x["id"]: x["id"] for x in instance["points"]}

            keypoints = []
            name = ""
            supercategory = ""
            id_ = 0
            try:
                template_names.add(str(instance["templateId"]))
                skeleton = __make_skeleton(instance)
                keypoints = __make_keypoints(instance)
                id_ = next(id_generator)
                if "classId" in instance.keys():
                    supercategory = instance["classId"]
                else:
                    supercategory = str(instance["templateId"])
                if "className" in instance.keys():
                    name = str(instance["className"])
                else:
                    name = str(instance["templateId"])
            except Exception as e:
                logger.error(e)

            category_item = {
                "name": name,
                "supercategory": supercategory,
                "skeleton": skeleton,
                "keypoints": keypoints,
                "id": id_,
            }
            categories.append(category_item)
        image_id = next(id_generator_img)
        image_info = make_image_info(
            json_["metadata"]["name"],
            json_["metadata"]["height"],
            json_["metadata"]["width"],
            image_id,
        )
        images.append(image_info)

        for instance in json_data:
            cat_id = None
            if instance["type"] == "template":
                for cat in categories:
                    if cat["name"] == instance["className"]:
                        cat_id = cat["supercategory"]

                annotation = __make_annotations(
                    instance, id_generator_anno, cat_id, image_info["id"]
                )
                annotations.append(annotation)
        images_converted.append(json_data)
    finish_event.set()
    tqdm_thread.join()
    return (categories, annotations, images)



"""

This object will receive the strategy from outside and will convert according to
   said strategy.

"""
from .coco_converters.coco_strategies import CocoKeypointDetectionStrategy
from .coco_converters.coco_strategies import CocoObjectDetectionStrategy
from .coco_converters.coco_strategies import CocoPanopticConverterStrategy
from .dataloop_converters.dataloop_strategies import DataLoopStrategy
from .googlecloud_converters.googlecloud_strategies import GoogleCloudStrategy
from .labelbox_converters.labelbox_strategies import LabelBoxStrategy
from .sagemaker_converters.sagemaker_strategies import SageMakerStrategy
from .supervisely_converters.supervisely_strategies import SuperviselyStrategy
from .vgg_converters.vgg_strategies import VGGStrategy
from .voc_converters.voc_strategies import VocStrategy
from .vott_converters.vott_strategies import VoTTStrategy
from .yolo_converters.yolo_strategies import YoloStrategy


class Converter:
    def __init__(self, args):
        self.output_dir = args.output_dir
        self._select_strategy(args)

    def convert_from_sa(self):
        self.strategy.sa_to_output_format()

    def convert_to_sa(self):
        self.strategy.to_sa_format()

    def __set_strategy(self, c_strategy):
        self.strategy = c_strategy

    def _select_strategy(self, args):
        if args.dataset_format == "COCO":
            if args.task == "instance_segmentation" or args.task == "object_detection":
                c_strategy = CocoObjectDetectionStrategy(args)
            if args.task == "keypoint_detection":
                c_strategy = CocoKeypointDetectionStrategy(args)
            if args.task == "panoptic_segmentation":
                c_strategy = CocoPanopticConverterStrategy(args)
        elif args.dataset_format == "DataLoop":
            c_strategy = DataLoopStrategy(args)
        elif args.dataset_format == "GoogleCloud":
            c_strategy = GoogleCloudStrategy(args)
        elif args.dataset_format == "LabelBox":
            c_strategy = LabelBoxStrategy(args)
        elif args.dataset_format == "SageMaker":
            c_strategy = SageMakerStrategy(args)
        elif args.dataset_format == "Supervisely":
            c_strategy = SuperviselyStrategy(args)
        elif args.dataset_format == "VGG":
            c_strategy = VGGStrategy(args)
        elif args.dataset_format == "VOC":
            c_strategy = VocStrategy(args)
        elif args.dataset_format == "VoTT":
            c_strategy = VoTTStrategy(args)
        elif args.dataset_format == "YOLO":
            if args.task == "object_detection":
                c_strategy = YoloStrategy(args)
        else:
            pass

        self.__set_strategy(c_strategy)



"""SuperAnnotate format annotation JSON helpers"""
import datetime
import json

from superannotate.lib.app.exceptions import AppException


def fill_in_missing(annotation_json, image_name: str = ""):
    for field in ["instances", "comments", "tags"]:
        if field not in annotation_json:
            annotation_json[field] = []
    if "metadata" not in annotation_json:
        annotation_json["metadata"] = {"name": image_name}


def _preprocess_annotation_json(annotation_json, image_name: str = ""):
    path = None
    if not isinstance(annotation_json, dict) and annotation_json is not None:
        path = annotation_json
        annotation_json = json.load(open(annotation_json))
    elif annotation_json is None:
        annotation_json = {}

    fill_in_missing(annotation_json, image_name)

    return (annotation_json, path)


def _postprocess_annotation_json(annotation_json, path):
    if path is not None:
        json.dump(annotation_json, open(path, "w"))
        return None
    else:
        return annotation_json


def _add_created_updated(annotation):
    created_at = (
        datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%f")[
            :-3
        ]
        + "Z"
    )
    annotation["createdAt"] = created_at
    annotation["updatedAt"] = created_at
    return annotation


def add_annotation_comment_to_json(
    annotation_json,
    comment_text,
    comment_coords,
    comment_author,
    resolved=False,
    image_name="",
):
    """Add a comment to SuperAnnotate format annotation JSON

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)

    :param comment_text: comment text
    :type comment_text: str

    :param comment_coords: [x, y] coords
    :type comment_coords: list

    :param comment_author: comment author email
    :type comment_author: str

    :param resolved: comment resolve status
    :type resolved: bool
    """

    if len(comment_coords) != 2:
        raise AppException("Comment should have two values")

    annotation_json, path = _preprocess_annotation_json(
        annotation_json, image_name=image_name
    )

    user_action = {"email": comment_author, "role": "Admin"}

    annotation = {
        "x": comment_coords[0],
        "y": comment_coords[1],
        "correspondence": [{"text": comment_text, "email": comment_author}],
        "resolved": resolved,
        "createdBy": user_action,
        "updatedBy": user_action,
    }
    annotation = _add_created_updated(annotation)
    annotation_json["comments"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_bbox_to_json(
    annotation_json,
    bbox,
    annotation_class_name,
    annotation_class_attributes=None,
    error=None,
    image_name: str = "",
):
    """Add a bounding box annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)

    :param bbox: 4 element list of top-left x,y and bottom-right x, y coordinates
    :type bbox: list of floats

    :param annotation_class_name: annotation class name
    :type annotation_class_name: str

    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts

    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """

    if len(bbox) != 4:
        raise AppException("Bounding boxes should have 4 float elements")

    annotation_json, path = _preprocess_annotation_json(annotation_json, image_name)

    annotation = {
        "type": "bbox",
        "points": {"x1": bbox[0], "y1": bbox[1], "x2": bbox[2], "y2": bbox[3]},
        "className": annotation_class_name,
        "error": error,
        "groupId": 0,
        "pointLabels": {},
        "locked": False,
        "visible": True,
        "attributes": []
        if annotation_class_attributes is None
        else annotation_class_attributes,
    }

    annotation = _add_created_updated(annotation)
    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)


def add_annotation_point_to_json(
    annotation_json,
    point,
    annotation_class_name,
    image_name,
    annotation_class_attributes=None,
    error=None,
):
    """Add a point annotation to SuperAnnotate format annotation JSON

    annotation_class_attributes has the form [ {"name" : "<attribute_value>", "groupName" : "<attribute_group>"},  ... ]

    :param annotation_json: annotations in SuperAnnotate format JSON or filepath to JSON
    :type annotation_json: dict or Pathlike (str or Path)
    :param point: [x,y] list of coordinates
    :type point: list of floats
    :param annotation_class_name: annotation class name
    :type annotation_class_name: str
    :param annotation_class_attributes: list of annotation class attributes
    :type annotation_class_attributes: list of 2 element dicts
    :param error: if not None, marks annotation as error (True) or no-error (False)
    :type error: bool
    """
    if len(point) != 2:
        raise AppException("Point should be 2 element float list.")

    annotation_json, path = _preprocess_annotation_json(annotation_json, image_name)

    annotation = {
        "type": "point",
        "x": point[0],
        "y": point[1],
        "className": annotation_class_name,
        "error": error,
        "groupId": 0,
        "pointLabels": {},
        "locked": False,
        "visible": True,
        "attributes": []
        if annotation_class_attributes is None
        else annotation_class_attributes,
    }

    annotation = _add_created_updated(annotation)
    annotation_json["instances"].append(annotation)

    return _postprocess_annotation_json(annotation_json, path)



"""
Labelbox to SA conversion method
"""
import threading

import cv2
from superannotate.logger import get_default_logger

from ....common import tqdm_converter
from ....common import write_to_json
from ..sa_json_helper import _create_sa_json
from ..sa_json_helper import _create_vector_instance
from .labelbox_helper import _create_attributes_list
from .labelbox_helper import _create_classes_id_map

logger = get_default_logger()


def labelbox_to_sa(json_data, output_dir, task):
    classes = _create_classes_id_map(json_data)
    if task == "object_detection":
        instance_types = ["bbox"]
    elif task == "instance_segmentation":
        instance_types = ["polygon"]
    elif task == "vector_annotation":
        instance_types = ["bbox", "polygon", "line", "point"]

    images_converted = []
    images_not_converted = []
    finish_event = threading.Event()
    tqdm_thread = threading.Thread(
        target=tqdm_converter,
        args=(len(json_data), images_converted, images_not_converted, finish_event),
        daemon=True,
    )
    logger.info("Converting to SuperAnnotate JSON format")
    tqdm_thread.start()
    for data in json_data:
        if "objects" not in data["Label"].keys():
            file_name = data["External ID"] + "___objects.json"
            write_to_json(
                output_dir / file_name,
                {"metadata": {}, "instances": [], "tags": [], "comments": []},
            )
            continue

        instances = data["Label"]["objects"]
        sa_instances = []

        for instance in instances:
            class_name = instance["value"]
            attributes = []
            if "classifications" in instance.keys():
                attributes = _create_attributes_list(instance["classifications"])

            lb_type = list(set(instance_types) & set(instance.keys()))
            if len(lb_type) != 1:
                continue

            if lb_type[0] == "bbox":
                points = (
                    instance["bbox"]["left"],
                    instance["bbox"]["top"],
                    instance["bbox"]["left"] + instance["bbox"]["width"],
                    instance["bbox"]["top"] + instance["bbox"]["height"],
                )
                instance_type = "bbox"
            elif lb_type[0] == "polygon":
                points = []
                for point in instance["polygon"]:
                    points.append(point["x"])
                    points.append(point["y"])
                instance_type = "polygon"
            elif lb_type[0] == "line":
                points = []
                for point in instance["line"]:
                    points.append(point["x"])
                    points.append(point["y"])
                instance_type = "polyline"
            elif lb_type[0] == "point":
                points = (instance["point"]["x"], instance["point"]["y"])
                instance_type = "point"

            sa_obj = _create_vector_instance(
                instance_type, points, {}, attributes, class_name
            )
            sa_instances.append(sa_obj)

        images_converted.append(data["External ID"])
        file_name = "%s___objects.json" % data["External ID"]
        try:
            img = cv2.imread(str(output_dir / data["External ID"]))
            H, W, _ = img.shape
        except Exception as e:
            logger.warning(
                "Can't open %s image. 'height' and 'width' for SA JSON metadata will set to zero",
                data["External ID"],
            )
            H, W = 0, 0

        sa_metadata = {"name": data["External ID"], "height": H, "width": W}
        sa_json = _create_sa_json(sa_instances, sa_metadata)
        write_to_json(output_dir / file_name, sa_json)

    finish_event.set()
    tqdm_thread.join()
    return classes



"""
"""
from superannotate.logger import get_default_logger

from .coco_converters.coco_to_sa_pixel import coco_instance_segmentation_to_sa_pixel
from .coco_converters.coco_to_sa_pixel import coco_panoptic_segmentation_to_sa_pixel
from .coco_converters.coco_to_sa_vector import coco_instance_segmentation_to_sa_vector
from .coco_converters.coco_to_sa_vector import coco_keypoint_detection_to_sa_vector
from .coco_converters.coco_to_sa_vector import coco_object_detection_to_sa_vector
from .coco_converters.sa_pixel_to_coco import sa_pixel_to_coco_instance_segmentation
from .coco_converters.sa_pixel_to_coco import sa_pixel_to_coco_panoptic_segmentation
from .coco_converters.sa_vector_to_coco import sa_vector_to_coco_instance_segmentation
from .coco_converters.sa_vector_to_coco import sa_vector_to_coco_keypoint_detection
from .coco_converters.sa_vector_to_coco import sa_vector_to_coco_object_detection
from .dataloop_converters.dataloop_to_sa_vector import dataloop_to_sa
from .googlecloud_converters.googlecloud_to_sa_vector import googlecloud_to_sa_vector
from .labelbox_converters.labelbox_to_sa_pixel import (
    labelbox_instance_segmentation_to_sa_pixel,
)
from .labelbox_converters.labelbox_to_sa_vector import labelbox_to_sa
from .sagemaker_converters.sagemaker_to_sa_pixel import (
    sagemaker_instance_segmentation_to_sa_pixel,
)
from .sagemaker_converters.sagemaker_to_sa_vector import (
    sagemaker_object_detection_to_sa_vector,
)
from .supervisely_converters.supervisely_to_sa_pixel import (
    supervisely_instance_segmentation_to_sa_pixel,
)
from .supervisely_converters.supervisely_to_sa_vector import (
    supervisely_keypoint_detection_to_sa_vector,
)
from .supervisely_converters.supervisely_to_sa_vector import supervisely_to_sa
from .vgg_converters.vgg_to_sa_vector import vgg_to_sa
from .voc_converters.voc_to_sa_pixel import voc_instance_segmentation_to_sa_pixel
from .voc_converters.voc_to_sa_vector import voc_instance_segmentation_to_sa_vector
from .voc_converters.voc_to_sa_vector import voc_object_detection_to_sa_vector
from .vott_converters.vott_to_sa_vector import vott_to_sa
from .yolo_converters.yolo_to_sa_vector import yolo_object_detection_to_sa_vector

logger = get_default_logger()

CONVERSION_ALGORITHMS = {
    "from": {
        "COCO": {
            "Vector": {
                "keypoint_detection": coco_keypoint_detection_to_sa_vector,
                "instance_segmentation": coco_instance_segmentation_to_sa_vector,
                "object_detection": coco_object_detection_to_sa_vector,
            },
            "Pixel": {
                "panoptic_segmentation": coco_panoptic_segmentation_to_sa_pixel,
                "instance_segmentation": coco_instance_segmentation_to_sa_pixel,
            },
        },
        "VOC": {
            "Vector": {
                "object_detection": voc_object_detection_to_sa_vector,
                "instance_segmentation": voc_instance_segmentation_to_sa_vector,
            },
            "Pixel": {"instance_segmentation": voc_instance_segmentation_to_sa_pixel},
        },
        "LabelBox": {
            "Vector": {
                "object_detection": labelbox_to_sa,
                "instance_segmentation": labelbox_to_sa,
                "vector_annotation": labelbox_to_sa,
            },
            "Pixel": {
                "instance_segmentation": labelbox_instance_segmentation_to_sa_pixel
            },
        },
        "DataLoop": {
            "Vector": {
                "object_detection": dataloop_to_sa,
                "instance_segmentation": dataloop_to_sa,
                "vector_annotation": dataloop_to_sa,
            },
            "Pixel": {},
        },
        "Supervisely": {
            "Vector": {
                "vector_annotation": supervisely_to_sa,
                "instance_segmentation": supervisely_to_sa,
                "object_detection": supervisely_to_sa,
                "keypoint_detection": supervisely_keypoint_detection_to_sa_vector,
            },
            "Pixel": {
                "instance_segmentation": supervisely_instance_segmentation_to_sa_pixel
            },
        },
        "VoTT": {
            "Vector": {
                "instance_segmentation": vott_to_sa,
                "object_detection": vott_to_sa,
                "vector_annotation": vott_to_sa,
            },
            "Pixel": {},
        },
        "SageMaker": {
            "Vector": {"object_detection": sagemaker_object_detection_to_sa_vector},
            "Pixel": {
                "instance_segmentation": sagemaker_instance_segmentation_to_sa_pixel
            },
        },
        "VGG": {
            "Vector": {
                "object_detection": vgg_to_sa,
                "instance_segmentation": vgg_to_sa,
                "vector_annotation": vgg_to_sa,
            },
            "Pixel": {},
        },
        "GoogleCloud": {
            "Vector": {"object_detection": googlecloud_to_sa_vector},
            "Pixel": {},
        },
        "YOLO": {
            "Vector": {"object_detection": yolo_object_detection_to_sa_vector},
            "Pixel": {},
        },
    },
    "to": {
        "COCO": {
            "Vector": {
                "instance_segmentation": sa_vector_to_coco_instance_segmentation,
                "object_detection": sa_vector_to_coco_object_detection,
                "keypoint_detection": sa_vector_to_coco_keypoint_detection,
            },
            "Pixel": {
                "panoptic_segmentation": sa_pixel_to_coco_panoptic_segmentation,
                "instance_segmentation": sa_pixel_to_coco_instance_segmentation,
            },
        },
    },
}


class baseStrategy:
    def __init__(self, args):
        if args.dataset_format not in ("COCO", "VOC"):
            logger.warning(
                "Beta feature. %s to SuperAnnotate annotation format converter is in BETA state."
                % args.dataset_format
            )
        self.project_type = args.project_type
        self.dataset_name = args.dataset_name
        self.export_root = args.export_root
        self.output_dir = args.output_dir
        self.task = args.task
        self.direction = args.direction
        self.conversion_algorithm = CONVERSION_ALGORITHMS[self.direction][
            args.dataset_format
        ][self.project_type][self.task]

        self.name = f"{args.dataset_format} {self.task} converter"

        self.failed_conversion_cnt = 0

    def __str__(self):
        return "%s object" % (self.name)

    def set_output_dir(self, output_dir_):
        self.output_dir = output_dir_

    def set_export_root(self, export_dir):
        self.export_root = export_dir

    def set_dataset_name(self, dname):
        self.dataset_name = dname

    def increase_converted_count(self):
        self.failed_conversion_cnt = self.failed_conversion_cnt + 1

    def set_num_converted(self, num_converted_):
        self.num_converted = num_converted_



from lib.core.usecases.annotations import *  # noqa: F403 F401
from lib.core.usecases.folders import *  # noqa: F403 F401
from lib.core.usecases.images import *  # noqa: F403 F401
from lib.core.usecases.integrations import *  # noqa: F403 F401
from lib.core.usecases.items import *  # noqa: F403 F401
from lib.core.usecases.models import *  # noqa: F403 F401
from lib.core.usecases.projects import *  # noqa: F403 F401
